import numpy as np
import ffmpeg
import subprocess
import sys, os, time
from typing import Any
if sys.platform != 'win32':
    import fcntl

class FrameWriter:
    '''Video frame writer to the output video file using ffmpeg.
    
    Args:
        output_file_name (str): output video file name
        size (tuple[int, int]): frame size (width, height)
        fps (int): frames per second
        bitrate (str): bit rate
        qmin (int): min video quantizer scale (VBR)
        qmax (int): max video quantizer scale (VBR)
        pix_fmt (str|tuple[str, str]): output pixel format or tuple of input/output pixel format
        audio (ffmpeg.nodes.FilterableStream): audio stream which is generated by ffmpeg.input
        stdout (bool): if true, writer object outputs the ffmpeg messages to stdout
    
    Attributes:
        video_file_name (str): output video file name
        frame (np.ndarray): Array of frame buffer which shape is (size[1], size[0], 3) and dtype is np.uint8
    '''
    
    video_file_name: str
    frame: np.ndarray
    
    def __init__(self, 
                 output_file_name: str, 
                 size: tuple[int, int] = (1280, 720), 
                 fps: int = 30, 
                 bitrate: str = '10240k', 
                 qmin: int = 2, 
                 qmax: int = 31, 
                 pix_fmt: str|tuple[str, str] = 'yuv420p', 
                 audio: ffmpeg.nodes.FilterableStream|None = None, 
                 stdout: bool = False):
        self.video_file_name = output_file_name
        self.fps = fps
        self.stdout = stdout
        video_width, video_height = size
        output_args: list[Any] = [self.video_file_name]
        if audio:
            output_args.insert(0, audio)
        if isinstance(pix_fmt, tuple) or isinstance(pix_fmt, list):
            input_pix_fmt, output_pix_fmt = pix_fmt
        elif isinstance(pix_fmt, str):
            input_pix_fmt, output_pix_fmt = 'rgb24', pix_fmt
        else:
            raise ValueError('Type of pix_fmt must be str or tuple[str, str]')
        if input_pix_fmt == 'rgb24':
            self.frame = np.zeros((video_height, video_width, 3), dtype=np.uint8)
        elif input_pix_fmt == 'rgb48':
            self.frame = np.zeros((video_height, video_width, 3), dtype=np.uint16)
        else:
            raise ValueError(f'Unsupported input pixel format: {input_pix_fmt}')
        if sys.platform == 'win32':
            self.stdout = False
        args = (
            ffmpeg
            .input('pipe:', format='rawvideo', pix_fmt=input_pix_fmt, r=fps, s=f'{video_width}x{video_height}')
            .filter_('colorspace', 'bt709', iall='bt601-6-625', fast='1')
            .output(*output_args, 
                    sws_flags='spline+accurate_rnd+full_chroma_int', 
                    color_range=1, colorspace=1, color_primaries=1, color_trc=1, 
                    pix_fmt=output_pix_fmt, video_bitrate=bitrate, qmin=qmin, qmax=qmax)
            .overwrite_output()
            .compile()
        )
        self.process = subprocess.Popen(
            args, 
            stdin=subprocess.PIPE, 
            stdout=subprocess.DEVNULL, 
            stderr=subprocess.PIPE if self.stdout else subprocess.DEVNULL
        )
        if self.stdout:
            fl =  fcntl.fcntl(self.process.stderr, fcntl.F_GETFL)
            fcntl.fcntl(self.process.stderr, fcntl.F_SETFL, fl|os.O_NONBLOCK)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def close(self):
        '''Close writer object.
        '''
        try:
            self.process.stdin.close()
            self.process.wait()
            self.process.stdout.close()
            self.process.stderr.close()
            if self.stdout:
                self.__print_ffmpeg_messages()
                print()
        except:
            pass

    def write_frame(self):
        '''Add a frame in the buffer of the writer object to the output video file.
        '''
        self.write(self.frame)

    def write(self, frame):
        '''Add a frame of the external buffer to the output video file.

        Args:
            frame (np.ndarray): an external buffer which shape is (size[1], size[0], 3) and dtype is np.uint8
        '''
        if frame.shape != self.frame.shape:
            raise ValueError(f'Shape of frame parameter should be {self.frame.shape}, but {frame.shape}')
        self.process.stdin.write(frame.tobytes())
        self.__print_ffmpeg_messages()
    
    def __print_ffmpeg_messages(self):
        if self.stdout and self.process.stderr.readable():
            data = self.process.stderr.read()
            if data:
                print(data.decode('utf-8'), end='')
            sys.stdout.flush()



import numpy as np
import ffmpeg
import subprocess
from dataclasses import dataclass
from typing import Union

@dataclass
class Probe:
    '''Probing first video stream from the input video file.

    Args:
        input_file_name (str): input video file name

    Attributes:
        video_file_name (str): input video file name
        width (int): width of video frame
        height (int): height of video frame
        sample_aspect_ratio (str): sample aspect ratio
        display_aspect_ratio (str): display aspect ratio
        duration (float): duration of video in seconds
        fps (float): frames per seconds
        n_frames (int): number of frames
        rotation (int): video rotation
    '''

    video_file_name: str
    width: int
    height: int
    sample_aspect_ratio: str
    display_aspect_ratio: str
    duration: float
    fps: float
    n_frames: int
    rotation: int

    def __init__(self, input_file_name: str):
        self.video_file_name = input_file_name
        probe = ffmpeg.probe(input_file_name)
        video_info = next(s for s in probe['streams'] if s['codec_type'] == 'video')
        self.width = int(video_info['width'])
        self.height = int(video_info['height'])
        self.sample_aspect_ratio = video_info['sample_aspect_ratio'] if 'sample_aspect_ratio' in video_info else None
        self.display_aspect_ratio = video_info['display_aspect_ratio'] if 'display_aspect_ratio' in video_info else None
        self.duration = float(video_info['duration'])
        self.fps = eval(video_info['r_frame_rate'])
        self.n_frames = int(eval('{0}*{1}'.format(video_info['duration'], video_info['r_frame_rate'])))
        self.rotation = 0
        if 'side_data_list' in video_info and len(video_info['side_data_list']):
            if 'rotation' in video_info['side_data_list'][0]:
                self.rotation = int(video_info['side_data_list'][0]['rotation'])


class FrameReader:
    '''Video frame reader from the input video file using ffmpeg.
    
    Args:
        input_file_name (str): input video file name
        ss (float): start time in seconds
        to (float): stop time in seconds, or end time if None is specified
        n_frames (int): number of frames
        filter_complex (dict[str, Union[dict, list, tuple]]): definition of a complex filtergraph
        pix_fmt (str): output pixel format: only "rgb24" (default) or "rgb48" supported

    Attributes:
        video_file_name (str): input video file name
        width (int): width of rotated video frame
        height (int): height of rotated video frame
        pix_fmt (str): output pixel format
    '''

    video_file_name: str
    width: int
    height: int

    def __init__(self, 
                 input_file_name: str, 
                 ss: float = 0, 
                 to: float = None, 
                 n_frames: int = None, 
                 filter_complex: dict[str, Union[dict, list, tuple]] = None, 
                 pix_fmt: str = 'rgb24'):
        if pix_fmt not in ['rgb24', 'rgb48']:
            raise  ValueError('pix_fmt must be "rgb24" or "rgb48"')
        self.pix_fmt = pix_fmt
        self.dtype, self.dbytes = (np.uint8, 1) if pix_fmt == 'rgb24' else (np.uint16, 2)
        self.video_file_name = input_file_name
        self.probe = Probe(input_file_name)
        if int(self.probe.rotation) % 180:
            self.width = self.probe.height
            self.height = self.probe.width
        else:
            self.width = self.probe.width
            self.height = self.probe.height
        if to is None:
            to = self.probe.duration
        process = ffmpeg.input(input_file_name, ss=ss, to=to)
        if n_frames is not None:
            process = process.filter('fps', n_frames / (to - ss))
        if filter_complex:
            for k, v in filter_complex.items():
                if type(v) == dict:
                    process = process.filter_(k, **v)
                elif type(v) == list or type(v) == tuple:
                    process = process.filter_(k, *v)
                else:
                    process = process.filter_(k, v)
        args = (
            process
            .output('pipe:', format='rawvideo', pix_fmt=pix_fmt)
            .compile()
        )
        self.process = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def close(self):
        '''Close reader object.
        '''
        try:
            self.process.stdout.close()
            self.process.wait()
        except:
            pass

    def read_frame(self):
        '''Read a frame from the input video file.
        
        Returns:
            np.ndarray: Array of frame buffer which shape is (height, width, 3) and dtype is np.uint8 (rgb24) or np.uint16 (rgb48)
        '''
        in_bytes = self.process.stdout.read(self.probe.width * self.probe.height * 3 * self.dbytes)
        if not in_bytes:
            return None
        return np.frombuffer(in_bytes, self.dtype).reshape([self.height, self.width, 3])

    def frames(self):
        '''Get a generator of input video frames.

        Returns:
            Generator[np.ndarray]: a generator of input video frames
        '''
        while (frame := self.read_frame()) is not None:
            yield frame
